Attribute VB_Name = "modSerialize"
Option Explicit

' Dieses Modul ist in der Lage, String-Arrays zu einzelnen Strings zu serialisieren, die dann
' beispielsweise ber Winsock verschickt werden knnen.
' Zu diesem Zweck verfgt jeder String ber einen Deskriptor mit drei Angaben:
'  - Die Lnge des gesamten Strings (abzglich 4 Byte)
'  - Die untere Arraygrenze
'  - Die obere Arraygrenze
' Anschlieend werden die einzelnen Elemente geschrieben, wobei jedes ber einen 4 Byte langen
' Deskriptor verfgt, der die Lnge des Elementes angibt.

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (ByRef Destination As Any, ByRef Source As Any, ByVal ByteLen As Long)

' Empfangsfenster. Hier werden alle eingehenden Daten zuerst gesammelt, bevor sie verwertet werden.
Private window As String

Private Sub WinsockReceiveData(ByVal data As String)
   ' Diese Prozedur simuliert den Empfang der Daten per Winsock.
   Dim n As Long                 ' Schleifenzhler fr die Auswertung
   Dim msg() As String           ' Die aktuelle Nachricht
   
   ' Empfangene Daten an das Empfangsfenster anhngen
   window = window & data
   
   ' Solange das Empfangsfenster mindestens eine komplette Nachricht enthlt:
   ' (Dies kann leicht berprft werden, da die ersten 4 Byte jeder Nachricht ihre Lnge angeben;
   '  Diese Lnge steht immer am Anfang des Empfangsfensters)
   Do While Len(window) - 4 >= StringToNumber(Left$(window, 4))
      ' Die Nachricht auslesen, den berstehenden "Rest" wieder in das Empfangsfenster geben
      msg = StringToStringArray(window, window)
      
      ' Die Nachricht irgendwie verwerten
      Call frmMain.Log("Es wurde eine Nachricht empfangen:")
      
      For n = LBound(msg) To UBound(msg)
         Call frmMain.Log(" " & CStr(n) & ": " & msg(n))
      Next n
      
      Call frmMain.Log("Ende der Nachricht")
   Loop
End Sub

Private Sub WinsockSendData(ByVal data As String)
   ' Diese Prozedur simuliert die Aufsplittung der Daten in kleinere Pakete, so wie es bei Winsock passieren wrde.
   ' Die zu sendenen Daten werden dabei in 128 Byte groe Pakete aufgeteilt und jedes Paket wird einzeln
   ' an die ReceiveData-Prozedur weitergeleitet.
   '
   ' In einem tatschlichen Programm muss man sich diese Mhe natrlich nicht machen ;)
   Dim start As Long: start = 1
   
   Do While start < Len(data)
      If Len(data) - start > 128 Then
         Call WinsockReceiveData(Mid$(data, start, 128))
         start = start + 128
      Else
         Call WinsockReceiveData(Mid$(data, start, Len(data) - start + 1))
         start = Len(data)
      End If
   Loop
End Sub

Public Sub WinsockTest()
   ' Eine Testprozedur, die zeigt, dass die Daten korrekt ausgegeben werden.
   Dim tmp() As String           ' Temporres String-Array
   Dim data As String            ' Zu sendende Daten
   
   tmp = Split("abc,def,ghi,jkl,mno,pqr,stu,vwx,yz", ",")
   data = StringArrayToString(tmp)
   tmp = Split("abcd,efgh,ijkl,mnop,qrst,uvwx,yz", ",")
   data = data & StringArrayToString(tmp)
   tmp = Split("1,2,3,4,5,6,7,8,9,0", ",")
   data = data & StringArrayToString(tmp)
   
   Call frmMain.Log("Testdaten: ")
   Call frmMain.Log(data)
   Call frmMain.Log("(Lnge: " & CStr(Len(data)) & "byte)")
   
   Call WinsockSendData(data)
End Sub

Public Sub WinsockTest2()
   ' Eine Testprozedur, die zufllig erstellte Strings sendet.
   Dim tmp(1 To 100) As String
   Dim data As String
   Dim n As Long
   Dim x As Long
   Dim length As Long
   
   For n = 1 To 100
      length = Rnd() * 100
      
      tmp(n) = Space(length)
      For x = 1 To length
         Mid$(tmp(n), x, 1) = Chr$(Rnd() * 255)
      Next x
   Next n
   
   data = StringArrayToString(tmp)
   
   Call frmMain.Log("Testdaten: ")
   Call frmMain.Log(data)
   Call frmMain.Log("(Lnge: " & CStr(Len(data)) & "byte)")
   
   Call WinsockSendData(data)
End Sub

Public Sub WinsockTest3(expr() As String)
   ' Eine Testprozedur, die Benutzerdefinierte Daten benutzt.
   Dim data As String
   
   data = StringArrayToString(expr)
   
   Call frmMain.Log("Testdaten: ")
   Call frmMain.Log(data)
   Call frmMain.Log("(Lnge: " & CStr(Len(data)) & "byte)")
   
   Call WinsockSendData(data)
End Sub

Public Function StringArrayToString(expr() As String) As String
   ' Diese Funktion serialisiert ein Stringarray, um es spter wieder deserialisieren zu knnen.
   Dim n As Long                 ' Schleifenzhler
   
   On Error GoTo err_SATS
   
   ' Arraygrenzen in den Deskriptor schreiben
   StringArrayToString = NumberToString(LBound(expr)) & NumberToString(UBound(expr))
   
   ' Alle Elemente schreiben
   For n = LBound(expr) To UBound(expr)
      '                                               Lngendeskriptor           Element
      StringArrayToString = StringArrayToString & NumberToString(Len(expr(n))) & expr(n)
   Next n
   
   ' Die Gesamtlnge des Strings vor den String an sich schreiben
   StringArrayToString = NumberToString(Len(StringArrayToString)) & StringArrayToString
   
   Exit Function
   
err_SATS:
   ' Im Fehlerfall einen vordefinierten String zurckgeben, der ein Array mit
   ' 1 leeren Eintrag codiert
   StringArrayToString = NumberToString(12) & NumberToString(0) & NumberToString(0) & NumberToString(0)
End Function

Public Function StringToStringArray(ByVal expr As String, Optional ByRef rest As String) As String()
   ' Diese Funktion deserialisiert einen String und gibt ein Stringarray zurck.
   
   Dim n As Long                 ' Schleifenzhler
   Dim start As Long             ' Lesezeiger (zeigt die aktuelle Position in expr an, die ausgewertet wird)
   Dim totalLength As Long       ' Totale Lnge des zu lesenden Strings (ohne Lngenangabe)
   Dim lb As Long, ub As Long    ' Arraygrenzen
   Dim length As Long            ' Lnge des aktuell ausgewerteten Elements
   Dim ret() As String           ' Temporres Array fr die Rckgabe
   
   ' Absolute Lnge auslesen (wichtig, um ggf. berstehenden Anteil abzuschneiden und zurckzugeben)
   totalLength = StringToNumber(Mid$(expr, 1, 4))
   ' Arraygrenzen auslesen
   lb = StringToNumber(Mid$(expr, 5, 4))
   ub = StringToNumber(Mid$(expr, 9, 4))
   
   ' Arraygrenzen festlegen
   ReDim ret(lb To ub)
   
   ' Lesezeiger auf den Beginn des Datenbereichs legen
   start = 13
   
   ' Alle Elemente auslesen
   For n = lb To ub
      ' Lnge des Elements auslesen
      length = StringToNumber(Mid$(expr, start, 4))
      ' Das Element selbst auslesen
      ret(n) = Mid$(expr, start + 4, length)
      
      ' Lesezeiger weiterschieben
      start = start + 4 + length
   Next n
   
   ' Den berstehenden Rest zurckgeben
   rest = Right$(expr, Len(expr) - 4 - totalLength)
   
   ' Das Array zurckgeben
   StringToStringArray = ret
End Function

Private Function NumberToString(ByVal expr As Long) As String
   Dim BA(3) As Byte             ' Das Bytearray, das die Zahl reprsentiert
   
   ' Den Long in das Bytearray kopieren
   Call CopyMemory(ByVal VarPtr(BA(0)), ByVal VarPtr(expr), 4)
   ' Das Bytearray in einen String konvertieren und zurckgeben
   NumberToString = StrConv(BA, vbUnicode)
End Function

Private Function StringToNumber(ByVal expr As String) As Long
   Dim BA() As Byte              ' Das Bytearray, das die Zahl reprsentiert
   
   ' Wenn ein leerer String bergeben wird, 0 zurckgeben
   If expr = "" Then
      StringToNumber = 0
      Exit Function
   End If

   ' Den String in ein Array aus Bytes konvertieren
   BA = StrConv(expr, vbFromUnicode)
   ' Die Bytes in einen Long kopieren
   Call CopyMemory(ByVal VarPtr(StringToNumber), ByVal VarPtr(BA(0)), 4)
End Function

